(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

module Make(V:Constant.S)(C:CompileCommon.Config) : XXXCompile.S =
  struct

    module ARM = ARMArch.Make(V)
    include CompileCommon.Make(C)(ARM)

    let do_sta = !Config.sta

(******)
    let ppo f k = k

    open ARM
    open C


(* Utilities *)
    let next_reg x = ARM.alloc_reg x

    let next_init st p init loc =
      let rec find_rec = function
        | (Reg (p0,r0),loc0)::_ when loc0 = loc && p = p0 ->
            r0,init,st
        | _::rem -> find_rec rem
        | [] ->
            let r = Symbolic_reg (sprintf "%s%i" loc p) in
            r,(Reg (p,r),loc)::init,st in
      find_rec init

    let pseudo = List.map (fun i -> Instruction i)


(*********)
(* loads *)
(*********)

    let emit_load st p init x =
      let rA,st = next_reg st in
      let rB,init,st = next_init st p init x in
      rA,init,pseudo [I_LDR (rA,rB)],st

    let emit_load_not_zero st p init x =
      let rA,st = next_reg st in
      let rB,init,st = next_init st p init x in
      let lab = Label.next_label "L" in
      rA,init,
      Label (lab,Nop)::pseudo [I_LDR (rA,rB); I_CMPI (rA,0); I_BNE (lab)],
      st

    let emit_load_not_eq st p init x rP =
      let rA,st = next_reg st in
      let rC,st = next_reg st in
      let rB,init,st = next_init st p init x in
      let lab = Label.next_label "L" in
      let out = Label.next_label "L" in
      rA,init,
      Instruction (I_MOVI (rC,200))::
      (* 200 X about 5 ins looks for a typical memory delay *)
      Label (lab,Nop)::
      pseudo
        [
         I_LDR (rA,rB); I_CMP (rA,rP) ;
         I_BNE out; I_ADD (DontSetFlags,rC,rC,-1) ;
         I_CMPI (rC,0) ; I_BNE lab ;
       ]@
      [Label (out,Nop)],
      st
        
    let emit_load_idx st p init x idx =
      let rA,st = next_reg st in
      let rB,init,st = next_init st p init x in
      rA,init,pseudo [I_LDR3 (rA,idx,rB)],st

(**********)
(* Stores *)
(**********)

    let emit_store_reg st p init x rA =
      let rB,init,st = next_init st p init x in
      init,[Instruction (I_STR (rA,rB))],st

    let emit_store_idx_reg st p init x idx rA =
      let rB,init,st = next_init st p init x in
      init,[Instruction (I_STR3 (rA,idx,rB))],st

    let emit_store st p init x v =
      let rA,st = next_reg st in
      let init,cs,st = emit_store_reg st p init x rA in
      init,Instruction (I_MOVI (rA,v))::cs,st

    let emit_store_idx st p init x idx v =
      let rA,st = next_reg st in
      let init,cs,st = emit_store_idx_reg st p init x idx rA in
      init,Instruction (I_MOVI (rA,v))::cs,st

(* No FNO yet *)
    let emit_fno  st p init x = assert false
    and emit_fno2  st p init x = assert false
    and emit_open_fno st p init x = assert false
    and emit_close_fno st p init lab r x = assert false

(*************)
(* Acccesses *)
(*************)

    let emit_access  st p init e = match e.dir with
    | R ->
        let r,init,cs,st = emit_load st p init e.loc in
        Some r,init,cs,st
    | W ->
        let init,cs,st = emit_store st p init e.loc e.v in
        None,init,cs,st

    let emit_access_dep_addr st p init e  r1 =
      let r2,st = next_reg st in
      let c =  I_XOR (DontSetFlags,r2,r1,r1) in
      match e.dir with
      | R ->
          let r,init,cs,st = emit_load_idx st p init e.loc r2 in
          Some r,init, Instruction c::cs,st
      | W ->
          let init,cs,st = emit_store_idx st p init e.loc r2 e.v in
          None,init,Instruction c::cs,st

    let emit_access_dep_data st p init e  r1 =
      match e.dir with
      | R ->Warn.fatal "data dependency to load"
      | W ->
          let r2,st = next_reg st in
          let cs2 =
            [Instruction (I_XOR (DontSetFlags,r2,r1,r1)) ;
             Instruction (I_ADD (DontSetFlags,r2,r2,e.v)) ; ] in
          let init,cs,st = emit_store_reg st p init e.loc r2 in
          None,init,cs2@cs,st

    let emit_access_ctrl isync st p init e r1 =      
      let lab = Label.next_label "LC" in
      let c =
        [Instruction (I_CMP (r1,r1));
         Instruction (I_BNE lab);
         Label (lab,Nop);] in
      match e.dir with
      | R ->
          let r,init,cs,st = emit_load st p init e.loc in
          Some r,init,c@cs,st
      | W ->
          let init,cs,st = emit_store st p init e.loc e.v in
          None,init,c@cs,st

    let emit_access_dep st p init e dp r1 = match dp with
    | ADDR -> emit_access_dep_addr st p init e r1
    | DATA -> emit_access_dep_data st p init e r1
    | CTRL -> emit_access_ctrl false st p init e r1

    let emit_fence f = Instruction (match f with | DMB -> I_DMB)

    let stronger_fence = DMB

  end
