(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Config
open Code

module type Compat = sig 
  val choice : check
end

module Make(Co:Compat)(C:XXXCompile.S) :
    sig
      val gen : ?relax:C.R.relax list -> ?safe:C.R.relax list -> int -> unit
    end

    =
  struct
    module D =
      DumpAll.Make
        (struct
          let name = !Config.name
          let canonical_only = !Config.cannonical_only
          let change_vars = false
          let outname = match !Config.tarfile with
          | None -> "."
          | Some s -> s
          module Tar =
            Tar.Make
              (struct
                let verbose = !Misc.verbose
                let outname = outname
              end)
        end)
        (C)
    open C.E
    open C.R

    let upto = !Config.upto

    module RelaxSet = C.R.Set

  let choice_sc (po_safe,fence_safe) e1 e2 = match e1,e2 with
(*
  Now reject all po;po, except for Rfi.
  A bit rude, maybe...
 *)
(* Assuming Dp is safe *)
    | Rf Int,Dp _| Dp _,Rf Int -> true
    | Dp _,Ws Int | Dp _,Fr Int ->        
        not (po_safe (dir_src e1) (dir_tgt e2))
    | Po _, Dp _ ->
        not (po_safe (dir_src e1) (dir_tgt e1)) &&
        not (po_safe (dir_src e1) (dir_tgt e2))
    | Dp _,Po _ ->
        not (po_safe (dir_src e2) (dir_tgt e2)) &&
        not (po_safe (dir_src e1) (dir_tgt e2))
(* Check Po is safe *)
    | Rf Int,Po _ ->
        po_safe (dir_src e2) (dir_tgt e2) &&
        not (po_safe (dir_src e1) (dir_tgt e2))
    | Po _,Rf Int ->
        po_safe (dir_src e1) (dir_tgt e1) &&
        not (po_safe (dir_src e1) (dir_tgt e2))
(* fenced s W R + Dp *)
    | Fenced(_,Same,_,Dir R), Dp (_,_,Dir R) ->
        not (fence_safe (dir_src e1) (dir_tgt e2))
(* Dp + fenced d R _ *)
    | Dp (_,_,Dir R),Fenced (_,Same,Dir R,_) ->
        not (fence_safe (dir_src e1) (dir_tgt e2))
(* Added *) 
    | _,_ ->
        match get_ie e1, get_ie e2 with
        | Int,Int -> false
        | Ext,_|_,Ext -> true 

(* Check altenance of com/po *)              
    let choice_critical _po_safe e1 e2 = match e1,e2 with
(* Two cases of allowed com composition *)
    | Ws Ext,Rf _
    | Fr Ext,Rf _ -> true
(* Otherwise require alternance *)
    | _,_ ->  C.E.get_ie e1 <> C.E.get_ie e2

  let choice_id _po_safe _e1 _e2 = true

  let choose c = match c with
    | Sc -> choice_sc
    | Critical -> choice_critical
    | Uni | Thin -> choice_id

    let pair_ok po_safe e1 e2 = match e1,e2 with
(*
  First reject some of hb' ; hb'
 *)

    | Hat,Hat   (* Hat *)
(* Ext Ext Only? *)
    | Ws _,Ws _ (* -> Ws *)
    | Ws _,Fr _ (* cannot occur *)
    | Fr _,Ws _ (* -> Fr*)
    | (Rf _|RfStar _),Fr _
(*    Rf _,Fr _ (* -> Ws *) May be interesting, because
      values are observed by outcome itself,
      also useful to add Fre after B-cumulativity *)
      -> false
(* Fence cumulativity *)
    | (RfStar _|Rf _),Fenced _
    | Fenced _,(Rf _|RfStar _) ->
        !Config.cumul && choose Co.choice po_safe e1 e2
    | _,_ -> choose Co.choice po_safe e1 e2

    let do_compat po_safe xs ys =
      let x = Misc.last xs
      and y = match ys with y::_ -> y | [] -> assert false in
      can_precede_dirs x y && pair_ok po_safe x y

    let can_precede po_safe (_,xs) k = match k with
    | [] -> true
    | (_,ys)::_ -> do_compat po_safe xs ys

    let pp_ess ess =
      String.concat " "
        (List.fold_right
           (fun (_,es) ->
             List.fold_right
               (fun e k -> pp_edge e::k)
               es)
           ess [])
    let edges_ofs rs = 
      List.map (fun r -> (r, edges_of r)) rs

(* Functional for recursive call of generators *)
    let sz r = match Co.choice with
    | Sc ->
        let _,es = r in
        List.length es
    | Critical|Thin|Uni -> 1

    let rec c_minprocs_es c = function
      | [] -> c
      | e::es ->
          let c = match get_ie e with
          | Int -> c
          | Ext -> c+1 in
          c_minprocs_es c es

    let rec c_minprocs_suff c = function
      | [] -> c
      | (_,es)::suff -> c_minprocs_suff (c_minprocs_es c es) suff
          
    let minprocs suff = c_minprocs_suff 0 suff

    let call_rec f0 po_safe n r suff f_rec k =
      if can_precede po_safe r suff && minprocs suff <= !Config.nprocs then
        let suff = r::suff
        and n = n-sz r in
        if !Misc.verbose > 2 then eprintf "CALL: %i %s\n%!" n (pp_ess suff) ;  
        let k =
          if (n = 0 || (n > 0 && upto)) then f0 po_safe suff k
          else k in
        if n <= 0 then k
        else f_rec n suff k
      else k

    module Dir2Set =
      MySet.Make
        (struct
          type t = extr * extr
          let compare = Pervasives.compare
        end)

    let extract_po rs =
      let d2 =
        List.fold_right
          (fun (r,_) k -> match r with
          | ERS [Po (_,e1,e2)] -> Dir2Set.add (e1,e2) k
          | _ -> k)
          rs Dir2Set.empty in
      fun e1 e2 -> Dir2Set.mem (e1,e2) d2

    let extract_fence rs =
      let d2 =
        List.fold_right
          (fun (r,_) k -> match r with
          | ERS [Fenced (_,_,e1,e2)] -> Dir2Set.add (e1,e2) k
          | _ -> k)
          rs Dir2Set.empty in
      fun e1 e2 -> Dir2Set.mem (e1,e2) d2

    let zyva relax safe n f =
      let relax = edges_ofs relax in
      let safe = edges_ofs safe in
      let po_safe = extract_po safe in
      let fence_safe = extract_fence safe in
      let po_safe = po_safe,fence_safe in

      let rec choose_relax rs k = match rs with
      | [] -> k
      | r0::rs -> (* Build simple cycles for relaxation r0 *)

          let call_rec = call_rec (f [fst r0]) po_safe  in

(* Add a safe edge to suffix *)        
          let rec add_safe ss n suf k =
            match ss with
            | [] -> k
            | s::ss ->
                let k = call_rec n s suf add_relaxs k in
                add_safe ss n suf k

(* Add some relax edges r0 to suffix, or nothing *)
          and add_relaxs n suf k = 
            let k = call_rec n r0 suf add_relaxs k in
            add_safe safe n suf k in

(* Start with a relax edge r0 *)          
          let k = call_rec n r0 [] add_relaxs k in
          choose_relax rs k in

(* Alternative: mix relaxation from relax list *)

      let all_relax k =
        let all = relax@safe
        and relax_set = RelaxSet.of_list (List.map fst relax) in
        let extract_relaxs suff =
          let suff_set = RelaxSet.of_list (List.map fst suff)  in
          RelaxSet.elements (RelaxSet.inter suff_set relax_set) in

        let call_rec =
          call_rec
            (fun po_safe suff k ->
              let rs = extract_relaxs suff in
              if List.length rs > !Config.max_relax then k
              else f rs po_safe suff k)
            po_safe in
        
(* Add a one edge to suffix *)        
        let rec add_one ss n suf k = match ss with
        | [] -> k
        | s::ss ->
            let k = call_rec n s suf (add_one all) k in
            add_one ss n suf k in

(* Force first edge to be a relaxed one *)
        let rec add_first rs k = match rs with
        | [] -> k
        | r::rs ->
            let k = call_rec n r [] (add_one all) k in
            add_first rs k in

        add_first relax k in
              
      

(* As a safety check, generate cycles with no relaxation *)
      let call_rec = call_rec (f []) po_safe in
      let rec no_relax ss n suf k = match ss with
      | [] -> k
      | s::ss ->
          let k = call_rec n s suf (no_relax safe) k in
          no_relax ss n suf k in

      fun k -> match relax with
      | [] -> no_relax safe n [] k
      | _  ->
          if !Config.mix && !Config.max_relax > 1 then
            all_relax k
          else
            choose_relax relax k

    let is_int e = match get_ie e with
    | Int -> true
    | Ext -> false

    let rec all_int l =
      match l with
      | [] -> true
      | [a] -> is_int a (* ??? pas trs utile :) *)
      | a::s -> (is_int a)&&(all_int s)

    let rec count_e ce = function
      | [] -> ce
      | e::es -> count_e (if is_int e then ce else ce+1) es


    let count_ext es = count_e 0 es 

    let change_loc e = match loc_sd e with
    | Same -> false
    | Diff -> true

    let count_p p =
      let rec do_rec c = function
        | [] -> c
        | x::xs -> do_rec (if p x then c+1 else c) xs in
      do_rec 0

    let count_changes = count_p change_loc

    let build_safe r0 es =
      let rs =
        List.fold_right (fun (r,_) -> RelaxSet.add r) es RelaxSet.empty in
      let rs = RelaxSet.diff rs (RelaxSet.of_list r0) in
      RelaxSet.elements rs

    let pp_rs rs =  sprintf "{%s}" (C.R.Set.pp_str "," C.R.pp_relax rs)

    let pp_rss rss = C.R.SetSet.pp_str " " pp_rs rss

    let last_check_call rset sset aset f rs po_safe res k =
      match res with
    | [] -> k
    | _ ->
        let lst = Misc.last res in
        if can_precede po_safe lst res then
          let es = List.map snd res in
          let le = List.flatten es in
          if (match Co.choice with | Sc -> true | Thin | Uni | Critical -> false) &&
            (count_ext le=1 || all_int le || count_changes le < 2) then k
          else begin
            let mk_info es =
              let ss = build_safe rs res in
(*
              let ass = C.R.relaxs_of aset es in
              let rss =
                C.R.SetSet.map (fun rs -> C.R.Set.inter rs rset) ass in
              let info =
                [
                 "Relax",pp_relax_list rs;
                 "Safe", pp_relax_list ss;
                 "Cycles",pp_rss ass ;
                 "Relaxs",pp_rss rss;
               ] in
*)  
              let info =
                [
                 "Relax",pp_relax_list rs;
                 "Safe", pp_relax_list ss;
               ] in
            info,C.R.Set.of_list rs in
            f le mk_info D.no_name k
          end
        else k

    let max_ins = !Config.max_ins 

    let last_minute ess =
      not (List.exists (fun es -> List.length es > max_ins) ess)

    let do_gen relax safe n =
      let sset = C.R.Set.of_list safe in
      let rset = C.R.Set.of_list relax in
      let aset = C.R.Set.union sset rset in
      D.all
        ~check:last_minute
        (fun f -> zyva relax safe n (last_check_call rset sset aset f))

    let debug_rs chan rs =
      List.iter (fun r -> fprintf chan "%s\n" (pp_relax r)) rs

    let secret_gen l1 l2 n =
      let l1 = expand_relaxs C.ppo l1
      and l2 = expand_relaxs C.ppo l2 in
      let s1 = C.R.Set.of_list l1
      and s2 = C.R.Set.of_list l2 in
      let l1 = C.R.Set.elements (C.R.Set.diff s1 s2) in
      if !Misc.verbose > 0 then begin
        eprintf "** Relax **\n" ;
        debug_rs stderr l1 ;
        eprintf "** Safe **\n" ;
        debug_rs stderr l2
      end ;
      do_gen l1 l2 n

(**********************)
(* Default edge lists *)
(**********************)

    let fold_ie f k = f (Int) (f (Ext) k)
    let fold_dir f k = f Irr k (* expand later ! *)
    let fold_dir2 f = fold_dir (fun i1 k -> fold_dir (f i1) k)
    let fold_sd f k = f (Same) (f Diff k)
    let fold_sd_dir f =
      fold_sd
        (fun sd -> fold_dir (fun d -> f sd d))
    let fold_sd_dir2 f = 
      fold_sd 
        (fun sd -> fold_dir2 (fun d1 d2 -> f sd d1 d2))
    let fold_all_fences f =
      fold_sd_dir2 (fun sd d1 d2 -> C.A.fold_all_fences (fun fe -> f fe sd d1 d2))
    let fold_cumul_fences f =
      fold_sd_dir2 (fun sd d1 d2 -> C.A.fold_cumul_fences (fun fe -> f fe sd d1 d2))
    let fold_cum f =
      fold_ie (fun ie -> fold_cumul_fences (fun fe sd d1 d2 -> f ie fe sd d1 d2)) 

    let er e = ERS [e]
    let safe =
      let k = [] in
      let k = fold_ie (fun ie k -> er (Ws ie)::er (Fr ie)::k) k in k

    let relax =
      let k = [] in
      let k = fold_dir2 (fun d1 d2 k -> er (Po (Diff, d1, d2))::k) k in
      let k = er (Po (Same, Dir R, Dir R))::k in
      let k = fold_all_fences (fun fe sd d1 d2 k -> er (Fenced (fe,sd,d1,d2))::k) k in
      let k =
        C.A.fold_dpr
          (fun dp k ->
            fold_sd (fun sd k -> er (Dp(dp,sd,Dir R))::k) k) k in
      let k =
        C.A.fold_dpw
          (fun dp k ->
            fold_sd (fun sd k -> er (Dp(dp,sd,Dir W))::k) k) k in
      let k = fold_ie (fun ie k -> er (Rf ie)::k) k in
      let k = fold_cum (fun ie fe sd d1 d2 k -> ac_fence fe sd d1 d2::k) k in
      let k = fold_cum (fun ie fe sd d1 d2 k -> bc_fence fe sd d1 d2::k) k in
      let k = er (Hat)::k in
      k
        
    let gen ?(relax=relax) ?(safe=safe) n = secret_gen relax safe n

  end
