(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Misc
open Printf

(* Configuration *)
let arch = ref PPC
let name = ref None
let tarfile = ref None
let rfi = ref false
let change_vars = ref false

let () = Config.nprocs := 1000

let opts =
  Config.common_specs @
  ("-name",
   Arg.String (fun n -> name := Some n),
   "<name> base name for tests")::
  ("-o",
   Arg.String (fun n -> tarfile := Some n),
   "<tar> output in archive <tar>")::
  ("-vars",Arg.Set change_vars,"special variable allocation")::
  Util.arch_opt (fun a -> arch := a) (pp_arch !arch)



module type Config = sig
  val name : string option
  val change_vars : bool
  module Tar : Tar.S
end

module Make (Config:Config) (C:XXXCompile.S) =
  struct
    module M = Top.Make(C)
    module D =
      DumpAll.Make
        (struct
          include Config
          let canonical_only = true
        end)
        (C)
    open Code
    open C.E

    let one_name = function
      | Po _ -> Some "po"
      | Fenced (f,_,_,_) -> Some (String.lowercase (C.A.pp_fence f))
      | Dp (dp,_,_) -> Some (String.lowercase (C.A.pp_dp dp))
      | Rf Int -> Some "rfi"
      | _ -> None

    let all_same = function
      | x::xs ->
          let rec do_rec = function
            | y::ys -> if x = y then do_rec ys else None
            | [] -> Some x in
          do_rec xs
      | [] -> None

            
    let mk_name base es =
      let name =
        let xs =
          List.fold_right
            (fun e r -> match one_name e with
            | Some n -> n::r
            | None -> r)
            es [] in
        let xs = match all_same xs,xs with
        | Some "po",_ -> []
        | Some x,[_] -> xs
        | Some x,_::_::_ -> [x ^ "s"]
        | None, _ -> xs
        | Some _,[] -> assert false in
        String.concat "+" (base::xs) in
(*
      eprintf "Edges [%s] -> %s\n"
        (C.E.pp_edges es) name ;
*)
      Some name

    let gen ess kont r =
      Misc.fold_cross ess
        (fun es r -> kont (List.flatten es) D.no_info mk_name r)
        r

    open Code

    let er e = C.R.ERS [e]

    let all_fences sd d1 d2 =
      C.A.fold_all_fences
        (fun f k -> er (C.E.Fenced (f,sd,Dir d1,Dir d2))::k)

    let allR sd d =
      let open C.E in
      er (Po (sd,Dir R,Dir d))::
      (match d with R -> C.A.fold_dpr | W -> C.A.fold_dpw)
        (fun dp k -> er (Dp (dp,sd,Dir d))::k)
        (all_fences sd R d [])      

    let allW sd d =
      let open C.E in
      er (Po (sd,Dir W,Dir d))::
      (all_fences sd W d [])      

    let parse_relaxs s = match s with
      | "allRR" -> allR Diff R
      | "allRW" -> allR Diff W
      | "allWR" -> allW Diff R
      | "allWW" -> allW Diff W
      | _ -> 
          let es = LexUtil.split s in
          List.map C.R.parse_relax es

    let parse_edges s =
      let rs = parse_relaxs s in
      List.fold_right (fun r k -> C.R.edges_of r :: k) rs []

      
      
    let zyva pp_rs =
      try
        let ess = List.map parse_edges pp_rs in
        D.all (gen ess)

      with Fatal msg -> Warn.warn_always "%s" msg
  end

let pp_es = ref []

let () =
  Util.parse_cmdline
    opts
    (fun x -> pp_es := x :: !pp_es)

let pp_es = List.rev !pp_es

let () =
  try
    let module Config = struct
    let name = !name
    let rfi = !rfi
    let change_vars = !change_vars
    let tarfile = match !tarfile with
    | None -> "."
    | Some s -> s
    module Tar =
      Tar.Make
        (struct
          let verbose = !Misc.verbose
          let outname = tarfile
        end)
  end in
    let module V = SymbConstant in
    let module C = struct let change_vars = !change_vars end in
    begin match !arch with
    | X86 ->
        let module M = Make(Config)(X86Compile.Make(V)(C)) in
        M.zyva
    | PPC ->
        let module M = Make(Config)(PPCCompile.Make(V)(C)) in
        M.zyva
    | ARM ->
        let module M = Make(Config)(ARMCompile.Make(V)(C)) in
        M.zyva
    end pp_es
  with
  | Misc.Exit -> ()
  | Misc.Fatal msg -> eprintf "Fatal error: %s\n" msg

