(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

(* Edges, ie specifications of an event pair in a model relation  *)

module type S = sig
  open Code

  type fence
  type dp

  val strong : fence

  val pp_fence : fence -> string

(* edge proper *)
  type edge =
    | Rf of ie | Fr of ie | Ws of ie
    | Po of sd*extr*extr | Fenced of fence*sd*extr*extr
    | Dp of dp*sd*extr
(* fancy *)
    | Hat
    | RfStar of ie

  val fold_edges : (edge -> 'a -> 'a) -> 'a -> 'a
  val iter_edges : (edge -> unit) -> unit

  val pp_edge : edge -> string

  val parse_edge : string -> edge
  val parse_edges : string -> edge list

  val pp_edges : edge list -> string

(* Get source and target event direction,
   Returning Irr means that a Read OR a Write is acceptable *)  
  val dir_src : edge -> extr 
  val dir_tgt : edge -> extr

(* Return edge with direction resolved *)
  val set_src : dir ->  edge -> edge
  val set_tgt : dir ->  edge -> edge

(* Does source and target events have the same or different locations? *)
  val loc_sd : edge -> sd

(* Internal (same proc) or external edge (different procs) *)
  val get_ie : edge -> ie

(* Can e1 target event direction be the same as e2 source event? *)
  val can_precede_dirs : edge -> edge -> bool
end

module Make(F:Fence.S) : S
with type fence = F.fence
and type dp = F.dp = struct

  open Code

  type fence = F.fence
  type dp = F.dp

  let pp_fence = F.pp_fence

  let strong = F.strong

(* edge proper *)
  type edge =
    | Rf of ie | Fr of ie | Ws of ie
    | Po of sd*extr*extr | Fenced of fence*sd*extr*extr
    | Dp of dp*sd*extr
(* fancy *)
    | Hat
    | RfStar of ie




open Printf
open Code
let pp_edge = function
  | Rf ie -> sprintf "Rf%s" (pp_ie ie)
  | RfStar ie -> sprintf "Rf*%s" (pp_ie ie)
  | Fr ie -> sprintf "Fr%s" (pp_ie ie)
  | Ws ie -> sprintf "Ws%s" (pp_ie ie)
  | Po (sd,e1,e2) -> sprintf "Po%s%s%s" (pp_sd sd) (pp_extr e1) (pp_extr e2)
  | Fenced (f,sd,e1,e2) -> 
      sprintf "%s%s%s%s" (F.pp_fence f) (pp_sd sd) (pp_extr e1) (pp_extr e2)
  | Dp (dp,sd,e) -> sprintf "Dp%s%s%s"
        (F.pp_dp dp) (pp_sd sd) (pp_extr e)
  | Hat -> "Hat"
 
let pp_strong sd e1 e2 = 
  sprintf "Fenced%s%s%s" (pp_sd sd) (pp_extr e1) (pp_extr e2)

(* Backward compatibility... *)

let pp_dp_default tag sd e = sprintf "%s%s%s" tag (pp_sd sd) (pp_extr e)

let fold_edges f r =
  let r = fold_ie (fun ie -> f (Rf ie)) r in
  let r = fold_ie (fun ie -> f (RfStar ie)) r in
  let r = fold_ie (fun ie -> f (Fr ie)) r in
  let r = fold_ie (fun ie -> f (Ws ie)) r in
  let r = fold_sd_extr_extr (fun sd e1 e2 r -> f (Po (sd,e1,e2)) r) r in
  let r =
    F.fold_all_fences
      (fun fe ->
        fold_sd_extr_extr
          (fun sd e1 e2 -> f (Fenced (fe,sd,e1,e2)))) r in
  let r =
    F.fold_dpr
      (fun dp -> fold_sd (fun sd -> f (Dp (dp,sd,Dir R)))) r in
  let r =
    F.fold_dpw
      (fun dp -> fold_sd (fun sd -> f (Dp (dp,sd,Dir W)))) r in
  let r = f Hat r in
  r

let iter_edges f = fold_edges (fun e () -> f e) ()

let t = Hashtbl.create 101

let add_lxm lxm e =  Hashtbl.add t lxm e

(* Fill up lexeme table *)
let () =
  iter_edges (fun e -> add_lxm (pp_edge e) e) ;
  fold_sd_extr_extr
    (fun sd e1 e2 () ->
      add_lxm
        (pp_strong sd e1 e2) (Fenced (F.strong,sd,e1,e2))) () ;
  let fill_opt tag dpo sd e = match dpo with
  | None -> ()
  | Some dp ->
      add_lxm
        (pp_dp_default tag sd e) 
        (Dp (dp,sd,e)) in
  fold_sd
    (fun sd () ->
      fill_opt "Dp" F.ddr_default sd (Dir R) ;
      fill_opt "Ctrl" F.ctrlr_default sd (Dir R) ;
      fill_opt "Dp" F.ddw_default sd (Dir W) ;
      fill_opt "Ctrl" F.ctrlw_default sd (Dir W) ;
      ()) () ;
  ()


let parse_edge s = 
  try Hashtbl.find t s
  with Not_found -> Warn.fatal "Bad edge: %s" s

let parse_edges s = List.map parse_edge (LexUtil.just_split s)

let parse_edge_opt s =
  try Some (Hashtbl.find t s)
  with Not_found -> None

let pp_edges es = String.concat " " (List.map pp_edge es)

let dir_tgt = function
  | Po(_,_,e)| Fenced(_,_,_,e)|Dp (_,_,e) -> e
  | Rf _| RfStar _ | Hat -> Dir R
  | Ws _|Fr _ -> Dir W

and dir_src = function
  | Po(_,e,_)| Fenced(_,_,e,_) -> e
  | Dp _|Fr _|Hat -> Dir R
  | Ws _|Rf _|RfStar _ -> Dir W

let set_tgt d e = match e with
  | Po(sd,src,_) -> Po (sd,src,Dir d)
  | Fenced(f,sd,src,_) -> Fenced(f,sd,src,Dir d)
  | Dp (dp,sd,_) -> Dp (dp,sd,Dir d) 
  | Rf _ |RfStar _ | Hat
  | Ws _|Fr _ -> e

and set_src d e = match e with
  | Po(sd,_,tgt) -> Po(sd,Dir d,tgt)
  | Fenced(f,sd,_,tgt) -> Fenced(f,sd,Dir d,tgt)
  | Fr _|Hat|Dp _
  | Ws _|Rf _|RfStar _ -> e


let loc_sd = function
  | Fr _|Ws _|Rf _|RfStar _|Hat -> Same
  | Po (sd,_,_) | Fenced (_,sd,_,_) | Dp (_,sd,_) -> sd

let get_ie = function
  | Po _|Dp _|Fenced _|Hat -> Int
  | Rf ie|RfStar ie|Fr ie|Ws ie -> ie

let can_precede_dirs  x y = match dir_tgt x,dir_src y with
    | Irr,Irr -> false
    | (Irr,Dir _) | (Dir _,Irr) -> true
    | Dir d1,Dir d2 -> d1=d2


end
