(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

module type S = sig
  type fence
  type dp
  type edge


  type relax =
(* Sequence of edges (eg Cumulativity) *)
    | ERS of edge list 
    | PPO

  val ac_fence : fence -> sd -> extr -> extr -> relax
  val bc_fence : fence -> sd -> extr -> extr -> relax
  val bc_dp : dp -> sd -> extr -> relax

(* Call function over all reckognized relaxations *)
  val fold_relax : (relax -> 'a -> 'a) -> 'a -> 'a

  val compare : relax -> relax -> int
  val pp_relax : relax -> string
  val pp_relax_list : relax list -> string
  val edges_of : relax -> edge list

(* Replace Irr directions in par expansion to W and R *)
  val expand_relaxs :
      ((relax -> relax list -> relax list) -> relax list -> relax list) ->
        relax list -> relax list

  val com : relax list
  val po : relax list

(* parsing *)
  val parse_relax : LexUtil.t -> relax
  val parse_relaxs : LexUtil.t list -> relax list

(* Sets *)
  module Set : MySet.S with type elt = relax
  val pp_set : out_channel -> Set.t -> unit

  (* Presence of cumulativity relaxations in a set *)
  val cumul_in : Set.t -> bool

  (* Remove cumulativity relaxations from set *)
  val remove_cumul : Set.t -> Set.t

  (* Expand cumulativity relaxations in set *)
  val expand_cumul : Set.t -> Set.t

  module SetSet : MySet.S with type elt = Set.t
  val pp_set_set : out_channel -> SetSet.t -> unit

  (* Apply expand cumul to all sets in a set of sets *)
  val expand_cumuls : SetSet.t -> SetSet.t

(* Map *)
  module Map : Map.S with type key = relax

(* From edge cycle to relax cycle *)
  val relaxs_of : Set.t -> edge list -> SetSet.t
end

module Make
    (F:Fence.S)
    (E:Edge.S with type fence = F.fence and type dp = F.dp) : S
with type fence = E.fence
and type dp = E.dp
and type edge = E.edge
= struct 
  type fence = E.fence
  type dp = E.dp
  type edge = E.edge

  type relax =
    | ERS of edge list
    | PPO

  let edges_of r = match r with
  | ERS es -> es
  | PPO -> assert false

  let compare r1 r2 = match r1,r2 with
  | PPO,PPO -> 0
  | PPO,ERS _ -> -1
  | ERS _,PPO -> 1
  | ERS l1,ERS l2 ->
      Pervasives.compare l1 l2


  open E

(* Cumulativity macros *)


  let ac_fence f sl d1 d2 = ERS [Rf Ext; Fenced (f,sl,d1,d2)]
  let bc_fence f sl d1 d2 = ERS [Fenced (f,sl,d1,d2); Rf Ext]
  let abc_fence f sl d1 d2 = ERS [Rf Ext; Fenced (f,sl,d1,d2); Rf Ext]
  let bc_dp dp sl d = ERS [Dp (dp,sl,d); Rf Ext]

(* Pretty print, macros are filtered and printed specially *)
  let pp_relax r =  match r with
  | ERS [e] -> E.pp_edge e
  | ERS [Rf Ext; Fenced _ as e] ->
      sprintf "AC%s" (pp_edge e)
  | ERS [Fenced _ as e; Rf Ext] ->
      sprintf "BC%s" (pp_edge e)
  | ERS [Rf Ext; Fenced _ as e; Rf Ext] ->
      sprintf "ABC%s" (pp_edge e)
  | ERS [Dp _ as e;Rf Ext] ->
      sprintf "BC%s" (pp_edge e) 
  | ERS es ->
      sprintf "[%s]" (String.concat "," (List.map pp_edge es))
  | PPO -> "PPO"


  let pp_relax_list lr = String.concat " " (List.map pp_relax lr)

(* Fold over all relaxations *)

  let fold_relax f k =
    let k = E.fold_edges (fun e -> f (ERS [e])) k in
    let k = 
      F.fold_cumul_fences
        (fun fe k ->
          let k =
            Code.fold_sd
              (fun sd k ->
                let k = f (abc_fence fe sd Irr Irr) k in
                f (abc_fence fe sd (Dir R) (Dir W)) k)
              k in
          Code.fold_sd_extr
            (fun sd e k ->
              let k = f (ac_fence fe sd Irr e) k in
              let k = f (ac_fence fe sd (Dir R) e) k in
              let k = f (bc_fence fe sd e Irr) k in
              f (bc_fence fe sd e (Dir W)) k)
            k) k in
    let k =
      F.fold_dpw
        (fun dpw k ->
          Code.fold_sd
            (fun sd k -> f (bc_dp dpw sd (Dir W)) k)
            k) k in
    let k = f PPO k in
    k

  let iter_relax f = fold_relax (fun r () -> f r) ()


(***********)
(* Parsing *)
(***********)

(*
  Same idea as for edges: pretty print all relaxations
  so as to build a table of reckognized relaxations.
 *)

(* Lexeme table *)
let t = Hashtbl.create 101

(* Fill up lexeme table *)
let () =
  iter_relax
    (fun e ->
      let pp = pp_relax e in
      Hashtbl.add t pp e);
  ()

  let do_parse_relax s = 
    try ERS [parse_edge s] (* Because some edges have special parsing *)
    with _ ->
      try Hashtbl.find t s
      with Not_found ->
        Warn.fatal "Bad relax: %s" s


(*************************************************************)
(* Expansion of irrelevant direction specifications in edges *)
(*************************************************************)

  let expand_dir d f = match d with
  | Dir _ -> f d
  | Irr -> fun k -> f (Dir W) (f (Dir R) k)
        
  let expand_dir2 e1 e2 f = 
    expand_dir e1
      (fun d1 -> expand_dir e2 (fun d2 -> f d1 d2))

  let do_expand_edge e f = match e with
  | Rf _ | RfStar _ | Fr _ | Ws _ | Hat |Dp _ -> f e
  | Po(sd,e1,e2) ->
      expand_dir2 e1 e2 (fun d1 d2 -> f (Po(sd,d1,d2)))
  | Fenced(fe,sd,e1,e2) ->
      expand_dir2 e1 e2 (fun d1 d2 -> f (Fenced(fe,sd,d1,d2)))


  let rec do_expand_edges es f suf = match es with
  | [] -> f suf
  | e::es ->
      do_expand_edge e
        (fun e k ->
          try
            let suf = match suf with
            | [] -> [e]
            | f::_ ->
                if can_precede_dirs e f then e::suf
                else raise Exit in
            do_expand_edges es f suf k
          with Exit -> k)

  let expand_edges es f = do_expand_edges (List.rev es) f []


(* Ouf, maintenat c'est facile *)
      

  let expand_relaxs ppo rs =

    let rec do_expand_relax r f = match r with
    | ERS es -> expand_edges es (fun es -> f (ERS es))
    | PPO  -> ppo (fun r -> do_expand_relax r f) in
    
    let expand_relax r =  do_expand_relax r Misc.cons in

    List.fold_right expand_relax rs []

  let er e = ERS [e]

  let com =
      [
       er (Rf Ext);
       er (Fr Ext);
       er (Ws Ext);
       ERS [Fr Ext ; Rf Ext;];
       ERS [Ws Ext; Rf Ext;];
     ]

  let po =
    er (Po (Diff,Irr,Irr))::
    F.fold_all_fences
      (fun f k ->
        er (Fenced (f,Diff,Irr,Irr))::
        (if F.orders f R R && not (F.orders f W R) then
          [ERS [Rf Int; Fenced (f,Diff,Dir R,Dir R)]]
        else [])@
        (if F.orders f R W && not (F.orders f W W) then
          [ERS [Rf Int; Fenced (f,Diff,Dir R,Dir W)]]
        else [])@k)
      []
      
      
  open LexUtil

  let parse_relax = function
    | One r -> do_parse_relax r
    | Seq [] -> Warn.fatal "Empty relaxation"
    | Seq es -> ERS (List.map E.parse_edge es)

  let parse_relaxs = List.map parse_relax

(********)
(* Sets *)
(********)

  module Set =
    MySet.Make
      (struct
        type t = relax
        let compare = compare
      end)

  let pp_set chan t =
    fprintf chan "{" ;
    Set.pp chan ", "
      (fun chan r -> fprintf chan "%s" (pp_relax r))
      t ;
    fprintf chan "}"    

  let is_cumul r = match r with
  | ERS [Rf Code.Ext;Fenced _;]
  | ERS [Fenced _; Rf Code.Ext;]
  | ERS [Rf Code.Ext; Fenced _; Rf Code.Ext;]
    -> true
  | _ -> false

  let cumul_in rs =  Set.exists is_cumul rs

  let remove_cumul rs = Set.filter (fun r -> not (is_cumul r)) rs

  let expand_cumul rs =
    let er e = ERS [e] in
    let xs = 
      Set.fold
        (fun r k ->
            match r with
          | ERS ([Rf Ext;Fenced _;] as rs)
          | ERS ([Fenced _; Rf Ext;] as rs)
          | ERS ([Rf Ext; Fenced _; Rf Ext;] as rs) ->
             Set.of_list (List.map er rs)::k
          | _ -> Set.singleton r::k)
        rs [] in
    Set.unions xs


  module SetSet = MySet.Make(Set)

  let pp_set_set chan ts = SetSet.pp chan " " pp_set ts

  let expand_cumuls rss =
    let xs =
      SetSet.fold
        (fun rs k -> expand_cumul rs::k)
        rss [] in
    SetSet.of_list  xs

(*********)

  module Map =
    Map.Make
      (struct
        type t = relax
        let compare = compare
      end)

    

(***************************************)
(* From edge cycle back to relaxations *)
(***************************************)

  let shift = function
    | [] -> assert false
    | x::xs -> xs @ [x]

  let rec match_edges ps es = match ps,es with
  | [],_ -> Some ([],es)
  | _::_,[] -> None
  | p::ps,e::es ->
      if p=e then match match_edges ps es with
      | Some (h,rem) -> Some (e::h,rem)
      | None -> None
      else None

  let rec match_head rs es =
    Set.fold
      (fun r k ->
        let ps = edges_of r in
        match match_edges ps es with
        | None -> k
        | Some (h,rem) ->
            List.fold_left
              (fun k rs -> (ERS h::rs)::k)
              k (matches rs rem))
      rs []

  and matches rs es = match es with
  | [] -> [[]]
  | _ -> match_head rs es

  let match_set rss = SetSet.of_list (List.map Set.of_list rss)

  let relaxs_of rs es =
    let rec do_rec k es =
      if k <= 0 then []
      else match_set (matches rs es)::do_rec (k-1) (shift es) in
  SetSet.unions (do_rec (List.length es) es)

end
