open Printf
open Globals

let gen_makefile arch = 
  let default_numthreads = match get_avail arch with
  | None -> 1
  | Some n -> n in
  let default_numruns    = get_run arch in
  let logdir     = "LOGDIR        ?= logs\n" in
  let numthreads = sprintf "NUMTHREADS    ?= %i\n" default_numthreads in 
  let numruns    = sprintf "NUMRUNS       ?= %i\n" default_numruns in
  let cpuinfo =
    match get_target_os arch with
    | Linux -> "\tcat /proc/cpuinfo > $(LOGDIR)/log_cpuinfo\n"
    | AIX ->
        "\techo 'CPUINFO unsupported for this OS' > $(LOGDIR)/log_cpuinfo\n"
    | Mac -> Warn.fatal "generateCrossDoc.ml does not support Mac"  in
  logdir ^
  numthreads ^
  numruns ^
  "\nall:\n" ^
  "\tmkdir -p $(LOGDIR)\n" ^
  "\t$(MAKE) $(MFLAGS) -C src\n" ^
  "\ttime sh -c \"cd src && sh run.sh -a $(NUMTHREADS) -r $(NUMRUNS)\\\n " ^
  "  > ../$(LOGDIR)/log_out 2> ../$(LOGDIR)/log_err\"\\\n" ^
  "  2> $(LOGDIR)/log_time\n" ^ 
  cpuinfo 

let gen_readme chan arch =
  let pl fmt = ksprintf (fun s -> fprintf chan "%s\n" s) fmt in
  let default_numthreads = match get_avail arch with
  | None -> 1
  | Some n -> n in
  let default_numruns = get_run arch in
  let default_size = get_size arch in
  pl   "This directory contains packaged files for running a set of" ;
  pl   "interesting litmus tests." ;
  pl "" ;
  pl "" ;
  pl "REQUIREMENTS:" ;
  pl "  unix shell (we use bash 3.1.17(1), most POSIX shells should work)" ;
  pl "  gcc       (we use gcc 4.1.2, any recent gcc should work)" ;
  pl "  GNU make   (we use GNU make 3.81)" ;  
  pl "  C sources and ggc options have been generated for: %s"
    (match get_target_os arch with
    | Linux -> "linux"
    | AIX   -> "AIX"
    | Mac   -> "MacOs") ;
  pl "" ;
  pl "" ;
  pl "INSTRUCTIONS:" ;
  pl "" ;  
  pl "- `make' at the top level will run tests, and produce 4 logfiles in" ;
  pl "   the LOGDIR subdirectory." ;
  pl "" ;
  pl "- make takes 3 arguments from shell:" ;
  pl "  LOGDIR (default logs): " ;
  pl "       Produced files are $(LOGDIR)/log_out, $(LOGDIR)/log_err" ;
  pl "       (which should usually be empty), $(LOGDIR)/log_cpuinfo (for" ;
  pl "       recording cpuinfo), and $(LOGDIR)/log_time (which records " ;
  pl "       total time elapsed). " ;
  pl "  NUMTHREADS (default %i):" default_numthreads ;
  pl "       The number of logical threads available. " ;
  pl "       The program attempts to run as many copies of the tests that would" ;
  pl "       fit in NUMTHREADS logical threads, eg NUMTHREADS=16 would allow us " ;
  pl "       to run 8 copies of 2-processor tests in parallel." ;
  begin match default_numthreads,get_affinity () with
  | 0, AffinityIncr _ ->
      pl "       The default value of 0 enables the automatic assignement" ;
      pl "       of NUMTHREAD to the number of logical processors of the" ;
      pl "       tested machine."
  | _,_ -> ()
  end ;
  pl "  NUMRUNS (default %i):" default_numruns ;
  pl "       Number of \"runs\", multiply by %i (in default configuration) for" 
       default_size ;
  pl "       number of test iterations run (thus default is %i iterations)."
       (default_size * default_numruns);
  pl "" ;
  pl "" ;
  pl "BEYOND DEFAULTS" ;
  pl "  The tests have been generated by litmus, see" ;
  pl "  <http://diy.inria.fr/doc/litmus.html>" ;
  ()                           


                   
